pub use ed25519_dalek::{Keypair, PublicKey, Signature, Signer};

use crate::prelude::*;

pub enum TrustLevel {
    Unknown,
    Distrusted,
    // Future extension
    // Transitively,
    Ultimately,
}

pub struct TrustDB {}

pub struct KeyExchangeInitiate {}

impl fleck_core::msg::MessageParams for KeyExchangeInitiate {
    const NAME: &'static str = "KeyExchangeInitiate";
    const ENCRYPTED: bool = false;
}

pub enum KeyExchange {
    WaitForResponse(x25519_dalek::PublicKey),
    Completed(SymmetricKey),
}

#[derive(serde::Serialize, serde::Deserialize, Default, Debug)]
pub enum PacketHeader {
    #[default]
    NoCrypto,
    Signed(Signature),
    // u64 here is the nonce
    Encrypted(u64),
}

pub struct SymmetricKey {
    key: [u8; 16],
    nonce: [u8; 16],
}

#[derive(Default)]
pub(crate) struct VerifyPacketSignature {}

impl Service for std::rc::Rc<VerifyPacketSignature> {
    fn setup(&self, api: &crate::Fleck) {
        api
            .channel::<fleck_core::ReceivePacketChannel>()
            .sub_opt(fleck_core::ReceiveOrder::Verify, self, VerifyPacketSignature::check);
    }
}

impl VerifyPacketSignature {
    fn check(&self, api: &crate::Fleck, mut message: fleck_core::msg::Message) -> Option<fleck_core::msg::Message> {
        match message.crypto_header {
            PacketHeader::Signed(signature) => {
                message.crypto_header = PacketHeader::NoCrypto;
                let to_verify = bincode::serialize(&message).expect("assume that we can re-serialize the message");

                // signature.verify(to_verify, signature);
                
                // let pubkey = api.services().with_service(|ns: &fleck_core::Nodes| { ns.self_node().pubkey() });

                // match pubkey {
                    // msg.crypto_header = PacketHeader::Signed(keypair.sign(&sign_data));
                // }
            }
            _ => {},
        }

        Some(message)
    }
}

#[derive(Default)]
pub(crate) struct SignPacket {}

impl Service for std::rc::Rc<SignPacket> {
    fn setup(&self, api: &crate::Fleck) {
        api
            .channel::<fleck_core::SendPacketChannel>()
            .sub_ref(fleck_core::SendOrder::Sign, self, SignPacket::sign);
    }
}

impl SignPacket {
    fn sign(&self, api: &crate::Fleck, message: &mut fleck_core::msg::Message) {
        let params = message
            .saved_params
            .expect("packet got to signing stage without saved parameters?");

        // don't do anything if the packet doesn't want a signature
        if !params.signed {
            return;
        }

        api.with_service(|ns: &fleck_core::NodeService| {
            let keypair = ns.self_node().keypair().unwrap();
            let sign_data = bincode::serialize(&message.content).unwrap();
            message.crypto_header = PacketHeader::Signed(keypair.sign(&sign_data));
        });
    }
}
