use std::rc::Rc;

pub mod fleck_core;
mod helper;
pub mod service;

pub mod prelude {
    pub use crate::fleck_core;
    pub use crate::helper::{AsAny, IntoWeak};
    pub use crate::service::{ChannelSpec, EventRoot, Service};
}

use prelude::*;

pub struct Fleck {
    services: service::ServiceStack,
}

impl Fleck {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            services: Default::default(),
        });

        res.add_service::<fleck_core::io::IOService>();
        res.add_service::<fleck_core::CoreInitService>();

        res
    }

    fn register_core_services(&self) {
        // core I/O service
        // initial outgoing/incoming/minor/major channels
        // self.services.borrow_mut().create_io_channels();

        // Node registration
        // self.add_service::<fleck_core::Nodes>();
        // Local node discovery
        // self.add_service::<fleck_core::lowlevel::LocalDiscovery>();
        // Actually sending packets
        // self.add_service::<fleck_core::lowlevel::SendPacket>();
        // Parsing incoming packets
        // self.add_service::<fleck_core::msg::ParsePacket>();
        // Signing packets
        // self.add_service::<crypto::SignPacket>();
    }

    pub fn run(&self) {
        self.with_service(|io: &fleck_core::io::IOService| {
            io.run(self);
        });
    }
}

impl Fleck {
    /// Add a service to the Fleck instance.
    pub fn add_service<S: Default + 'static>(&self)
    where
        Rc<S>: Service,
    {
        self.services.give_service::<S>(self);
    }
    /// Access a pre-existing service
    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        self.services.with_service(f)
    }
    /// Create an event channel from a specification
    pub fn create_channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, Self, CS::Data, CS::Priority>> {
        self.services.event_root().create_channel::<CS>();
        self.channel::<CS>()
    }
    /// Get an existing channel
    pub fn channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, Self, CS::Data, CS::Priority>> {
        self.services.event_root().channel::<CS>()
    }
    /// Queue a message onto a channel
    pub fn queue<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .event_root()
            .channel::<CS>()
            .as_ref()
            .queue(data);
    }
}
