use std::cell::RefCell;
use std::{
    collections::{HashMap, VecDeque},
    rc::{Rc, Weak},
};

pub enum SubscriptionFunction<Host: 'static, Context: 'static + ?Sized, Data: 'static> {
    ByRef(Box<dyn Fn(&Host, &Context, &mut Data)>),
    ByValue(Box<dyn Fn(&Host, &Context, Data) -> Option<Data>>),
}

impl<Host: 'static, Context: 'static + ?Sized, Data: 'static>
    SubscriptionFunction<Host, Context, Data>
{
    fn invoke(&self, host: &Host, context: &Context, mut data: Data) -> Option<Data> {
        match self {
            Self::ByRef(f) => {
                f(host, context, &mut data);
                Some(data)
            }
            Self::ByValue(f) => f(host, context, data),
        }
    }
}

struct ConcreteEventSub<Host: 'static, Context: 'static + ?Sized, Data: 'static> {
    host: Weak<Host>,
    callback: SubscriptionFunction<Host, Context, Data>,
}

trait EventSub<Context: 'static + ?Sized, Data: 'static> {
    fn is_healthy(&self) -> bool;
    fn invoke(&self, context: &Context, data: Data) -> Option<Data>;
}

impl<Host: 'static, Context: 'static + ?Sized, Data: 'static> EventSub<Context, Data>
    for ConcreteEventSub<Host, Context, Data>
{
    fn is_healthy(&self) -> bool {
        self.host.strong_count() > 0
    }
    fn invoke(&self, context: &Context, data: Data) -> Option<Data> {
        self.host
            .upgrade()
            .map(|h| self.callback.invoke(h.as_ref(), context, data))
            .flatten()
    }
}

pub struct Channel<Tag: 'static, Context: 'static + ?Sized, Data: 'static, Priority: 'static> {
    subs: RefCell<Vec<(Priority, Rc<dyn EventSub<Context, Data>>)>>,

    queue: RefCell<VecDeque<Data>>,

    _ghost: std::marker::PhantomData<(Tag, Priority)>,
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static, Priority: 'static> Default
    for Channel<Tag, Context, Data, Priority>
{
    fn default() -> Self {
        Self {
            subs: RefCell::new(Default::default()),
            queue: RefCell::new(Default::default()),
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static>
    Channel<Tag, Context, Data, NoPriorityTag>
{
    pub fn sub_ref<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, &mut Data) + 'static,
    >(
        &self,
        who: HC,
        cb: CB,
    ) {
        let sub = Rc::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: SubscriptionFunction::ByRef(Box::new(cb)),
        });

        self.subs.borrow_mut().push((NoPriorityTag {}, sub));
    }

    pub fn sub_opt<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, Data) -> Option<Data> + 'static,
    >(
        &self,
        who: HC,
        cb: CB,
    ) {
        let sub = Rc::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: SubscriptionFunction::ByValue(Box::new(cb)),
        });

        self.subs.borrow_mut().push((NoPriorityTag {}, sub));
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static, Priority: PartialOrd>
    Channel<Tag, Context, Data, Priority>
{
    pub fn sub_ref<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, &mut Data) + 'static,
    >(
        &self,
        p: Priority,
        who: HC,
        cb: CB,
    ) {
        let sub = Rc::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: SubscriptionFunction::ByRef(Box::new(cb)),
        });

        let mut subs = self.subs.borrow_mut();
        subs.push((p.into(), sub));
        // XXX: what happens if we actually get an undefined order?...
        subs.sort_by(|a, b| a.0.partial_cmp(&b.0).unwrap());
    }

    pub fn sub_opt<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, Data) -> Option<Data> + 'static,
    >(
        &self,
        p: Priority,
        who: HC,
        cb: CB,
    ) {
        let sub = Rc::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: SubscriptionFunction::ByValue(Box::new(cb)),
        });

        let mut subs = self.subs.borrow_mut();
        subs.push((p.into(), sub));
        // XXX: what happens if we actually get an undefined order?...
        subs.sort_by(|a, b| a.0.partial_cmp(&b.0).unwrap());
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static, Priority: 'static>
    Channel<Tag, Context, Data, Priority>
{
    pub fn queue(&self, data: Data) {
        self.queue.borrow_mut().push_back(data);
    }

    fn do_fire_all(&self, context: &Context) -> usize {
        let mut count = 0;

        while !self.queue.borrow().is_empty() {
            let mut event = self.queue.borrow_mut().pop_front();

            for sub in self.subs.borrow().iter() {
                event = sub.1.invoke(context, event.unwrap());
                if event.is_none() {
                    break;
                }
            }

            count += 1;
        }

        count
    }
}

trait ChannelMetadata<Context: 'static + ?Sized> {
    fn queue_len(&self) -> usize;
    fn fire_all(&self, context: &Context) -> usize;
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static, Priority: 'static>
    ChannelMetadata<Context> for Channel<Tag, Context, Data, Priority>
{
    fn queue_len(&self) -> usize {
        self.queue.borrow().len()
    }

    fn fire_all(&self, context: &Context) -> usize {
        self.do_fire_all(context)
    }
}

pub trait ChannelSpec: 'static {
    type Tag: 'static;
    type Data: 'static;
    type Priority: 'static + Clone;
}

#[derive(Clone)]
pub struct NoPriorityTag;

impl<Tag: 'static, Data: 'static, Priority: 'static + Clone + PartialOrd> ChannelSpec
    for (Tag, Data, Priority)
{
    type Tag = Tag;
    type Data = Data;
    type Priority = Priority;
}

impl<Tag: 'static, Data: 'static> ChannelSpec for (Tag, Data) {
    type Tag = Tag;
    type Data = Data;
    type Priority = NoPriorityTag;
}

pub struct EventRoot<Context: 'static + ?Sized> {
    pub(crate) channels: RefCell<HashMap<std::any::TypeId, Rc<dyn std::any::Any>>>,
    metadata: RefCell<HashMap<std::any::TypeId, (&'static str, Rc<dyn ChannelMetadata<Context>>)>>,
    _ghost: std::marker::PhantomData<Context>,
}

impl<Context: 'static + ?Sized> Default for EventRoot<Context> {
    fn default() -> Self {
        Self {
            channels: Default::default(),
            metadata: Default::default(),
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<Context: 'static + ?Sized> EventRoot<Context> {
    pub fn create_channel<CS: ChannelSpec>(&self) {
        let tid = std::any::TypeId::of::<Channel<CS::Tag, Context, CS::Data, CS::Priority>>();

        let ch = Rc::new(Channel::<CS::Tag, Context, CS::Data, CS::Priority>::default());
        self.metadata
            .borrow_mut()
            .insert(tid, (std::any::type_name::<CS>(), ch.clone()));
        self.channels.borrow_mut().insert(tid, ch);
    }

    pub fn channel<'l, 's: 'l, CS: ChannelSpec>(
        &'s self,
    ) -> Rc<Channel<CS::Tag, Context, CS::Data, CS::Priority>> {
        let tid = std::any::TypeId::of::<Channel<CS::Tag, Context, CS::Data, CS::Priority>>();

        let own = self.channels.borrow();

        if own.contains_key(&tid) {
            let entry = own.get(&tid).unwrap().clone();

            entry
                .downcast::<Channel<CS::Tag, Context, CS::Data, CS::Priority>>()
                .expect("internal inconsistency?")
        } else {
            panic!(
                "Asked for channel {} that has not been created!",
                std::any::type_name::<CS>()
            )
        }
    }

    pub(crate) fn fire_all(&self, context: &Context) {
        let mut any = true;
        while any {
            any = false;

            for ch in self.metadata.borrow().iter() {
                let count = ch.1 .1.fire_all(context);
                if count > 0 {
                    log::trace!("Queue {} processed {} event(s)", ch.1 .0, count);
                }
                any = any || count > 0;
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use std::ops::Deref;
    use std::{
        cell::{Cell, RefCell},
        rc::Rc,
    };

    use super::EventRoot;

    struct Receiver {
        int_count: Cell<usize>,
    }

    impl Receiver {
        fn receive_i32(&self, _ctx: &(), _val: &mut i32) {
            self.int_count.set(self.int_count.get() + 1);
        }
    }

    struct OrderedReceiver {
        id: i32,
        order: Rc<RefCell<Vec<i32>>>,
    }

    impl OrderedReceiver {
        fn receive_i32(&self, _ctx: &(), _val: &mut i32) {
            self.order.borrow_mut().push(self.id)
        }
    }

    struct IntTag;
    type IntChannel = (IntTag, i32);

    #[derive(PartialEq, PartialOrd, Clone)]
    enum IntPriority {
        First,
        Second,
        Third,
    }
    type IntPriorityChannel = (IntTag, i32, IntPriority);

    #[test]
    fn simple_fire() {
        let mut root = EventRoot::default();

        root.create_channel::<IntChannel>();

        let recv = Rc::new(Receiver {
            int_count: Cell::new(0),
        });

        root.channel::<IntChannel>()
            .sub_ref(&recv, Receiver::receive_i32);
        root.channel::<IntChannel>().queue(0i32);
        assert_eq!(recv.int_count.get(), 0);
        root.channel::<IntChannel>().do_fire_all(&());
        assert_eq!(recv.int_count.get(), 1);
    }

    #[test]
    fn priority_fire() {
        let mut root = EventRoot::default();

        root.create_channel::<IntPriorityChannel>();

        let order = Rc::new(RefCell::new(Vec::new()));

        let recv1 = Rc::new(OrderedReceiver {
            id: 1,
            order: order.clone(),
        });
        let recv2 = Rc::new(OrderedReceiver {
            id: 2,
            order: order.clone(),
        });
        let recv3 = Rc::new(OrderedReceiver {
            id: 3,
            order: order.clone(),
        });

        root.channel::<IntPriorityChannel>().sub_ref(
            IntPriority::Second,
            &recv2,
            OrderedReceiver::receive_i32,
        );
        root.channel::<IntPriorityChannel>().sub_ref(
            IntPriority::First,
            &recv1,
            OrderedReceiver::receive_i32,
        );
        root.channel::<IntPriorityChannel>().sub_ref(
            IntPriority::Third,
            &recv3,
            OrderedReceiver::receive_i32,
        );
        root.channel::<IntPriorityChannel>().queue(0i32);
        assert_eq!(order.borrow().deref(), &vec![]);
        root.channel::<IntPriorityChannel>().do_fire_all(&());
        assert_eq!(order.borrow().deref(), &vec![1, 2, 3]);
    }
}
